from metadata_basetests import BaseTests

from pointcyto.data.metadata_classification import MetaDataClassification


class TestMetaDataClassification(BaseTests):
    metadata = MetaDataClassification
